# -*- coding: utf-8 -*-

# MIT license
#
# Copyright (C) 2015 by XESS Corp.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import division
from __future__ import absolute_import
from builtins import str
from past.utils import old_div
import sys
import math
import re
from copy import copy
from pprint import pprint
import importlib
from affine import Affine
from .common import *

__all__ = ['kipart']  # Only export this routine for use by the outside world.

THIS_MODULE = sys.modules[__name__]  # Ref to this module for making named calls.

# Settings for creating the KiCad schematic part symbol.
# Dimensions are given in mils (0.001").

# Origin point.
XO = 0
YO = 0

# Pin settings.
PIN_LENGTH = 200
PIN_SPACING = 100
PIN_NUM_SIZE = 50  # Font size for pin numbers.
PIN_NAME_SIZE = 50  # Font size for pin names.
PIN_NAME_OFFSET = 40  # Separation between pin and pin name.
PIN_ORIENTATION = 'left'
PIN_STYLE = 'line'
SHOW_PIN_NUMBER = True  # Show pin numbers when True.
SHOW_PIN_NAME = True  # Show pin names when True.
SINGLE_PIN_SUFFIX = ''
MULTI_PIN_SUFFIX = '*'
PIN_SPACER_PREFIX = '*'

# Settings for box drawn around pins in a unit.
BOX_LINE_WIDTH = 12
FILL = 'no_fill'

# Part reference.
REF_SIZE = 60  # Font size.
REF_Y_OFFSET = 250

# Part number.
PART_NUM_SIZE = 60  # Font size.
PART_NUM_Y_OFFSET = 150

# Part footprint
PART_FOOTPRINT_SIZE = 60  # Font size.
PART_FOOTPRINT_Y_OFFSET = 50

# Part manufacturer number.
PART_MPN_SIZE = 60  # Font size.
PART_MPN_Y_OFFSET = -50

# Mapping from understandable pin orientation name to the orientation
# indicator used in the KiCad part library. This mapping looks backward,
# but if pins are placed on the left side of the symbol, you actually
# want to use the pin symbol where the line points to the right.
# The same goes for the other sides.
PIN_ORIENTATIONS = {
    '': 'R',
    'left': 'R',
    'right': 'L',
    'bottom': 'U',
    'down': 'U',
    'top': 'D',
    'up': 'D',
}
scrubber = re.compile('[^\w~#]+')
PIN_ORIENTATIONS = {
    scrubber.sub('', k).lower(): v
    for k, v in list(PIN_ORIENTATIONS.items())
}

ROTATION = {'left': 0, 'right': 180, 'bottom': 90, 'top': -90}

# Mapping from understandable pin type name to the type
# indicator used in the KiCad part library.
PIN_TYPES = {
    'input': 'I',
    'inp': 'I',
    'in': 'I',
    'clk': 'I',
    'output': 'O',
    'outp': 'O',
    'out': 'O',
    'bidirectional': 'B',
    'bidir': 'B',
    'bi': 'B',
    'inout': 'B',
    'io': 'B',
    'iop': 'B',
    'tristate': 'T',
    'tri': 'T',
    'passive': 'P',
    'pass': 'P',
    'unspecified': 'U',
    'un': 'U',
    '': 'U',
    'analog': 'U',
    'power_in': 'W',
    'pwr_in': 'W',
    'pwrin': 'W',
    'power': 'W',
    'pwr': 'W',
    'ground': 'W',
    'gnd': 'W',
    'power_out': 'w',
    'pwr_out': 'w',
    'pwrout': 'w',
    'pwr_o': 'w',
    'open_collector': 'C',
    'opencollector': 'C',
    'open_coll': 'C',
    'opencoll': 'C',
    'oc': 'C',
    'open_emitter': 'E',
    'openemitter': 'E',
    'open_emit': 'E',
    'openemit': 'E',
    'oe': 'E',
    'no_connect': 'N',
    'noconnect': 'N',
    'no_conn': 'N',
    'noconn': 'N',
    'nc': 'N',
}
PIN_TYPES = {scrubber.sub('', k).lower(): v for k, v in list(PIN_TYPES.items())}

# Mapping from understandable pin drawing style to the style
# indicator used in the KiCad part library.
PIN_STYLES = {
    'line': '',
    '': '',
    'inverted': 'I',
    'inv': 'I',
    '~': 'I',
    '#': 'I',
    'clock': 'C',
    'clk': 'C',
    'rising_clk': 'C',
    'inverted_clock': 'IC',
    'inv_clk': 'IC',
    'clk_b': 'IC',
    'clk_n': 'IC',
    '~clk': 'IC',
    '#clk': 'IC',
    'input_low': 'L',
    'inp_low': 'L',
    'in_lw': 'L',
    'in_b': 'L',
    'in_n': 'L',
    '~in': 'L',
    '#in': 'L',
    'clock_low': 'CL',
    'clk_low': 'CL',
    'clk_lw': 'CL',
    'output_low': 'V',
    'outp_low': 'V',
    'out_lw': 'V',
    'out_b': 'V',
    'out_n': 'V',
    '~out': 'V',
    '#out': 'V',
    'falling_edge_clock': 'F',
    'falling_clk': 'F',
    'fall_clk': 'F',
    'non_logic': 'X',
    'nl': 'X',
    'analog': 'X',
}
PIN_STYLES = {scrubber.sub('', k).lower(): v for k, v in list(PIN_STYLES.items())}

# Mapping from understandable box fill-type name to the fill-type
# indicator used in the KiCad part library.
FILLS = {'no_fill': 'N', 'fg_fill': 'F', 'bg_fill': 'f'}

# Format strings for various items in a KiCad part library.
LIB_HEADER = 'EESchema-LIBRARY Version 2.3\n'
START_DEF = 'DEF {name} {ref} 0 {pin_name_offset} {show_pin_number} {show_pin_name} {num_units} L N\n'
END_DEF = 'ENDDEF\n'
REF_FIELD = 'F0 "{ref_prefix}" {x} {y} {ref_size} H V {horiz_just} CNN\n'
PART_FIELD = 'F1 "{part_num}" {x} {y} {ref_size} H V {horiz_just} CNN\n'
FOOTPRINT_FIELD = 'F2 "{footprint}" {x} {y} {ref_size} H I {horiz_just} CNN\n'
MPN_FIELD = 'F4 "{manf_num}" {x} {y} {ref_size} H I {horiz_just} CNN "manf#"\n'

START_DRAW = 'DRAW\n'
END_DRAW = 'ENDDRAW\n'
BOX = 'S {x0} {y0} {x1} {y1} {unit_num} 1 {line_width} {fill}\n'
PIN = 'X {name} {num} {x} {y} {length} {orientation} {num_sz} {name_sz} {unit_num} 1 {pin_type} {pin_style}\n'


def annotate_pins(unit_pins):
    '''Annotate pin names to indicate special information.'''
    for name, pins in unit_pins:
        # If there are multiple pins with the same name in a unit, then append a
        # distinctive suffix to the pin name to indicate multiple pins are placed
        # at a single location on the unit. (This is done so multiple pins that
        # should be on the same net (e.g. GND) can be connected using a single
        # net connection in the schematic.)
        name_suffix = SINGLE_PIN_SUFFIX
        if len(pins) > 1:
            #name_suffix = MULTI_PIN_SUFFIX
            name_suffix = '[{}]'.format(len(pins))
        for pin in pins:
            pin.name += name_suffix


def get_pin_num_and_spacer(pin):
    pin_num = str(pin.num)
    pin_spacer = 0
    # spacer pins have pin numbers starting with a special prefix char.
    if pin_num.startswith(PIN_SPACER_PREFIX):
        pin_spacer = 1
        pin_num = pin_num[1:] # Remove the spacer prefix.
    return pin_num, pin_spacer


def count_pin_slots(unit_pins):
    '''Count the number of vertical pin slots needed for a column of pins.'''

    # Compute the # of slots for the column of pins, taking spacers into account.
    num_slots = 0
    pin_num_len = 0
    for name, pins in unit_pins:
        pin_spacer = 0
        pin_num_len = 0
        for pin in pins:
            pin_num, pin_spacer = get_pin_num_and_spacer(pin)
            pin_num_len = max(pin_num_len, len(pin_num))
        num_slots += pin_spacer # Add a slot if there was a spacer.
        # Add a slot if the pin number was more than just a spacer prefix.
        if pin_num_len > 0:
            num_slots += 1
    return num_slots


def pins_bbox(unit_pins):
    '''Return the bounding box of a column of pins and their names.'''

    if len(unit_pins) == 0:
        return [[XO, YO], [XO, YO]]  # No pins, so no bounding box.

    width = 0
    for name, pins in unit_pins:

        # Update the maximum observed width of a pin name. This is used later to
        # size the width of the box surrounding the pin names for this unit.
        width = max(width, len(pins[0].name) * PIN_NAME_SIZE)

    # Add the separation space before and after the pin name.
    width += PIN_LENGTH + 2 * PIN_NAME_OFFSET

    # Make bounding box an integer number of pin spaces so pin connections are always on the grid.
    width = math.ceil(old_div(float(width), PIN_SPACING)) * PIN_SPACING

    # Compute the height of the column of pins.
    height = count_pin_slots(unit_pins) * PIN_SPACING

    return [[XO, YO + PIN_SPACING], [XO + width, YO - height]]


def balance_bboxes(bboxes):
    '''Make the symbol more balanced by adjusting the bounding boxes of the pins on each side.'''
    X = 0
    Y = 1
    def find_bbox_bbox(*bboxes):
        '''Find the bounding box for a set of bounding boxes.'''
        bb = [[0,0], [0,0]]
        for bbox in bboxes:
            bb[0][X] = min(bb[0][X], bbox[0][X])
            bb[1][X] = max(bb[1][X], bbox[1][X])
            bb[0][Y] = max(bb[0][Y], bbox[0][Y])
            bb[1][Y] = min(bb[1][Y], bbox[1][Y])
        return bb

    # Determine the number of sides of the symbol with pins.
    num_sides = len(bboxes)

    if num_sides == 4:
        # If the symbol has pins on all four sides, then check to see if there
        # are approximately the same number of pins on all four sides. If so,
        # then equalize the bounding box for each side. Otherwise, equalize
        # the left & right bounding boxes and the top & bottom bounding boxes.
        lr_bbox = find_bbox_bbox(bboxes['left'], bboxes['right'])
        lr_hgt = abs(lr_bbox[0][Y] - lr_bbox[1][Y])
        tb_bbox = find_bbox_bbox(bboxes['top'], bboxes['bottom'])
        tb_hgt = abs(tb_bbox[0][Y] - tb_bbox[1][Y])
        if 0.75 <= float(lr_hgt)/float(tb_hgt) <= 1/0.75:
            bal_bbox = find_bbox_bbox(*list(bboxes.values()))
            for side in bboxes:
                bboxes[side] = copy(bal_bbox)
        else:
            bboxes['left'] = copy(lr_bbox)
            bboxes['right'] = copy(lr_bbox)
            bboxes['top'] = copy(tb_bbox)
            bboxes['bottom'] = copy(tb_bbox)
    elif num_sides == 3:
        # If the symbol only has pins on threee sides, then equalize the
        # bounding boxes for the pins on opposite sides and leave the
        # bounding box on the other side unchanged.
        if 'left' not in bboxes or 'right' not in bboxes:
            # Top & bottom side pins, but the left or right side is empty.
            bal_bbox = find_bbox_bbox(bboxes['top'], bboxes['bottom'])
            bboxes['top'] = copy(bal_bbox)
            bboxes['bottom'] = copy(bal_bbox)
        elif 'top' not in bboxes or 'bottom' not in bboxes:
            # Left & right side pins, but the top or bottom side is empty.
            bal_bbox = find_bbox_bbox(bboxes['left'], bboxes['right'])
            bboxes['left'] = copy(bal_bbox)
            bboxes['right'] = copy(bal_bbox)
    elif num_sides == 2:
        # If the symbol only has pins on two opposing sides, then equalize the
        # height of the bounding boxes for each side. Leave the width unchanged.
        if 'left' in bboxes and 'right' in bboxes:
            bal_bbox = find_bbox_bbox(bboxes['left'], bboxes['right'])
            bboxes['left'][0][Y] = bal_bbox[0][Y]
            bboxes['left'][1][Y] = bal_bbox[1][Y]
            bboxes['right'][0][Y] = bal_bbox[0][Y]
            bboxes['right'][1][Y] = bal_bbox[1][Y]
        elif 'top' in bboxes and 'bottom' in bboxes:
            bal_bbox = find_bbox_bbox(bboxes['top'], bboxes['bottom'])
            bboxes['top'][0][Y] = bal_bbox[0][Y]
            bboxes['top'][1][Y] = bal_bbox[1][Y]
            bboxes['bottom'][0][Y] = bal_bbox[0][Y]
            bboxes['bottom'][1][Y] = bal_bbox[1][Y]


def draw_pins(unit_num, unit_pins, bbox, transform, fuzzy_match):
    '''Draw a column of pins rotated/translated by the transform matrix.'''

    # String to add pin definitions to.
    pin_defn = ''

    # Find the actual height of the column of pins and subtract it from the
    # bounding box (which should be at least as large). Half the difference
    # will be the offset needed to center the pins on the side of the symbol.
    Y = 1 # Index for Y coordinate.
    pins_bb = pins_bbox(unit_pins)
    height_offset = abs(bbox[0][Y]-bbox[1][Y]) - abs(pins_bb[0][Y]-pins_bb[1][Y])
    height_offset /= 2
    height_offset -= height_offset % PIN_SPACING # Keep everything on the PIN_SPACING grid.

    # Start drawing pins from the origin.
    x = XO
    y = YO - height_offset

    for name, pins in unit_pins:

        # Detect pins with "spacer" pin numbers.
        pin_spacer = 0
        pin_num_len = 0
        for pin in pins:
            pin_num, pin_spacer = get_pin_num_and_spacer(pin)
            pin_num_len = max(pin_num_len, len(pin_num))
        y -= pin_spacer * PIN_SPACING # Add space between pins if there was a spacer.
        if pin_num_len == 0:
            continue # Omit pin if it only had a spacer prefix and no actual pin number.

        # Rotate/translate the current drawing point.
        (draw_x, draw_y) = transform * (x, y)

        # Use approximate matching to determine the pin's type, style and orientation.
        pin_type = find_closest_match(pins[0].type, PIN_TYPES, fuzzy_match)
        pin_style = find_closest_match(pins[0].style, PIN_STYLES, fuzzy_match)
        pin_side = find_closest_match(pins[0].side, PIN_ORIENTATIONS,
                                      fuzzy_match)

        if pins[0].hidden.lower().strip() in ['y', 'yes', 't', 'true', '1']:
            pin_style = 'N' + pin_style

        # Create all the pins with a particular name. If there are more than one,
        # they are laid on top of each other and only the first is visible.
        num_size = PIN_NUM_SIZE  # First pin will be visible.
        for pin in pins:

            pin_num = str(pin.num)

            # Remove any spacer prefix on the pin numbers.
            if pin_num.startswith(PIN_SPACER_PREFIX):
                pin_num = pin_num[1:]

            # Create a pin using the pin data.
            pin_defn += PIN.format(name=pin.name,
                                      num=pin_num,
                                      x=int(draw_x),
                                      y=int(draw_y),
                                      length=PIN_LENGTH,
                                      orientation=pin_side,
                                      num_sz=num_size,
                                      name_sz=PIN_NAME_SIZE,
                                      unit_num=unit_num,
                                      pin_type=pin_type,
                                      pin_style=pin_style)

            # Turn off visibility after the first pin.
            num_size = 0

        # Move to the next pin placement location on this unit.
        y -= PIN_SPACING

    return pin_defn  # Return part symbol definition with pins added.

def zero_pad_nums(s):
    # Pad all numbers in the string with leading 0's.
    # Thus, 'A10' and 'A2' will become 'A00010' and 'A00002' and A2 will
    # appear before A10 in a list.
    try:
        return re.sub(r'\d+', lambda mtch: '0' * (8 - len(mtch.group(0))) + mtch.group(0), s)
    except TypeError:
        return s # The input is probably not a string, so just return it unchanged.


def num_key(pin):
    '''Generate a key from a pin's number so they are sorted by position on the package.'''

    # Pad all numeric strings in the pin name with leading 0's.
    # Thus, 'A10' and 'A2' will become 'A00010' and 'A00002' and A2 will
    # appear before A10 in a list.
    return zero_pad_nums(pin[1][0].num)


def name_key(pin):
    '''Generate a key from a pin's name so they are sorted more logically.'''

    # Pad all numeric strings in the pin name with leading 0's.
    # Thus, 'adc10' and 'adc2' will become 'adc00010' and 'adc00002' and adc2 will
    # appear before adc10 in a list.
    return zero_pad_nums(pin[1][0].name)


def row_key(pin):
    '''Generate a key from the order the pins were entered into the CSV file.'''
    return pin[1][0].index


def draw_symbol(part_num, part_ref_prefix, part_footprint, part_manf_num, pin_data, sort_type, reverse, fuzzy_match):
    '''Add a symbol for a part to the library.'''

    # Start the part definition with the header.
    part_defn = START_DEF.format(name=part_num,
                    ref=part_ref_prefix,
                    pin_name_offset=PIN_NAME_OFFSET,
                    show_pin_number=SHOW_PIN_NUMBER and 'Y' or 'N',
                    show_pin_name=SHOW_PIN_NAME and 'Y' or 'N',
                    num_units=len(pin_data))

    # Determine if there are pins across the top of the symbol.
    # If so, right-justify the reference and part number so they don't
    # run into the top pins. If not, stick with left-justification.
    horiz_just = 'L'
    horiz_offset = PIN_LENGTH
    for unit in list(pin_data.values()):
        if 'top' in list(unit.keys()):
            horiz_just = 'R'
            horiz_offset = PIN_LENGTH - 50
            break

    # Create the field that stores the part reference.
    if part_ref_prefix:
        part_defn += REF_FIELD.format(ref_prefix=part_ref_prefix,
                                    x=XO + horiz_offset,
                                    y=YO + REF_Y_OFFSET,
                                    horiz_just=horiz_just,
                                    ref_size=REF_SIZE)

    # Create the field that stores the part number.
    if part_num:
        part_defn += PART_FIELD.format(part_num=part_num,
                                     x=XO + horiz_offset,
                                     y=YO + PART_NUM_Y_OFFSET,
                                     horiz_just=horiz_just,
                                     ref_size=PART_NUM_SIZE)

    # Create the field that stores the part footprint.
    if part_footprint:
        part_defn += FOOTPRINT_FIELD.format(footprint=part_footprint,
                                     x=XO + horiz_offset,
                                     y=YO + PART_FOOTPRINT_Y_OFFSET,
                                     horiz_just=horiz_just,
                                     ref_size=PART_FOOTPRINT_SIZE)

    # Create the field that stores the manufacturer part number.
    if part_manf_num:
        part_defn += MPN_FIELD.format(manf_num=part_manf_num,
                                     x=XO + horiz_offset,
                                     y=YO + PART_MPN_Y_OFFSET,
                                     horiz_just=horiz_just,
                                     ref_size=PART_MPN_SIZE)

    # Start the section of the part definition that holds the part's units.
    part_defn += START_DRAW

    # Get a reference to the sort-key generation function for pins.
    pin_key_func = getattr(THIS_MODULE, '{}_key'.format(sort_type))

    # This is the sort-key generation function for unit names.
    unit_key_func = lambda x: zero_pad_nums(x[0])

    # Now create the units that make up the part. Unit numbers go from 1
    # up to the number of units in the part. The units are sorted by their
    # names before assigning unit numbers.
    for unit_num, unit in enumerate([p[1] for p in sorted(pin_data.items(),key=unit_key_func)], 1):

        # The indices of the X and Y coordinates in a list of point coords.
        X = 0
        Y = 1

        # Initialize data structures that store info for each side of a schematic symbol unit.
        all_sides = ['left', 'right', 'top', 'bottom']
        bbox = {side: [(XO, YO), (XO, YO)] for side in all_sides}
        box_pt = {
            side: [XO + PIN_LENGTH, YO + PIN_SPACING]
            for side in all_sides
        }
        anchor_pt = {
            side: [XO + PIN_LENGTH, YO + PIN_SPACING]
            for side in all_sides
        }
        transform = {}

        # Annotate the pins for each side of the symbol.
        for side_pins in list(unit.values()):
            annotate_pins(list(side_pins.items()))

        # Determine the actual bounding box for each side.
        bbox = {}
        for side, side_pins in list(unit.items()):
            bbox[side] = pins_bbox(list(side_pins.items()))

        # Adjust the sizes of the bboxes to make the unit look more symmetrical.
        balance_bboxes(bbox)

        # Determine some important points for each side of pins.
        for side in unit:
            #
            #     C     B-------A
            #           |       |
            #     ------| name1 |
            #           |       |
            #     ------| name2 |
            #
            # A = anchor point = upper-right corner of bounding box.
            # B = box point = upper-left corner of bounding box + pin length.
            # C = upper-left corner of bounding box.
            anchor_pt[side] = [max(bbox[side][0][X], bbox[side][1][X]),
                               max(bbox[side][0][Y], bbox[side][1][Y])]
            box_pt[side] = [
                min(bbox[side][0][X], bbox[side][1][X]) + PIN_LENGTH,
                max(bbox[side][0][Y], bbox[side][1][Y])
            ]

        # AL = left-side anchor point.
        # AB = bottom-side anchor point.
        # AR = right-side anchor point.
        # AT = top-side anchor-point.
        #        +-------------+
        #        |             |
        #        |     TOP     |
        #        |             |
        # +------AL------------AT
        # |      |
        # |      |             +---------+
        # |      |             |         |
        # |  L   |             |         |
        # |  E   |             |    R    |
        # |  F   |             |    I    |
        # |  T   |             |    G    |
        # |      |             |    H    |
        # |      |             |    T    |
        # |      |             |         |
        # +------AB-------+    AR--------+
        #        | BOTTOM |
        #        +--------+
        #

        # Create zero-sized bounding boxes for any sides of the unit without pins.
        # This makes it simpler to do the width/height calculation that follows.
        for side in all_sides:
            if side not in bbox:
                bbox[side] = [(XO,YO), (XO,YO)]

        # This is the width and height of the box in the middle of the pins on each side.
        box_width = max(abs(bbox['top'][0][Y] - bbox['top'][1][Y]),
                        abs(bbox['bottom'][0][Y] - bbox['bottom'][1][Y]))
        box_height = max(abs(bbox['left'][0][Y] - bbox['left'][1][Y]),
                         abs(bbox['right'][0][Y] - bbox['right'][1][Y]))

        for side in all_sides:
            # Each side of pins starts off with the orientation of a left-hand side of pins.
            # Transformation matrix starts by rotating the side of pins.
            transform[side] = Affine.rotation(ROTATION[side])
            # Now rotate the anchor point to see where it goes.
            rot_anchor_pt = transform[side] * anchor_pt[side]
            # Translate the rotated anchor point to coincide with the AL anchor point.
            translate_x = anchor_pt['left'][X] - rot_anchor_pt[X]
            translate_y = anchor_pt['left'][Y] - rot_anchor_pt[Y]
            # Make additional translation to bring the AL point to the correct position.
            if side == 'right':
                # Translate AL to AR.
                translate_x += box_width
                translate_y -= box_height
            elif side == 'bottom':
                # Translate AL to AB
                translate_y -= box_height
            elif side == 'top':
                # Translate AL to AT
                translate_x += box_width
            # Create the complete transformation matrix = rotation followed by translation.
            transform[side] = Affine.translation(translate_x,
                                                 translate_y) * transform[side]
            # Also translate the point on each side that defines the box around the symbol.
            box_pt[side] = transform[side] * box_pt[side]

        # Draw the transformed pins for each side of the symbol.
        for side, side_pins in list(unit.items()):
            # If the pins are ordered by their row in the spreadsheet or by their name,
            # then reverse their order on the right and top sides so they go from top-to-bottom
            # on the right side and left-to-right on the top side instead of the opposite
            # as happens with counter-clockwise pin-number ordering.
            side_reverse = reverse
            if sort_type in ['name', 'row'] and side in ['right', 'top']:
                side_reverse = not reverse
            # Sort the pins for the desired order: row-wise, numeric (pin #), alphabetical (pin name).
            sorted_side_pins = sorted(list(side_pins.items()), key=pin_key_func, reverse=side_reverse)
            # Draw the transformed pins for this side of the symbol.
            part_defn += draw_pins(unit_num, sorted_side_pins, 
                                   bbox[side], transform[side], fuzzy_match)

            # Create the box around the unit's pins.
        part_defn += BOX.format(x0=int(box_pt['left'][X]),
                                  y0=int(box_pt['top'][Y]),
                                  x1=int(box_pt['right'][X]),
                                  y1=int(box_pt['bottom'][Y]),
                                  unit_num=unit_num,
                                  line_width=BOX_LINE_WIDTH,
                                  fill=FILLS[FILL])

    # Close the section that holds the part's units.
    part_defn += END_DRAW

    # Close the part definition.
    part_defn += END_DEF

    # Return complete part symbol definition.
    return part_defn


def is_pwr(pin, fuzzy_match):
    '''Return true if this is a power input pin.'''
    return find_closest_match(name=pin.type, name_dict=PIN_TYPES, fuzzy_match=fuzzy_match) == 'W'


def is_nc(pin, fuzzy_match):
    '''Return true if this is a no-connect pin.'''
    return find_closest_match(name=pin.type, name_dict=PIN_TYPES, fuzzy_match=fuzzy_match) == 'N'


def do_bundling(pin_data, bundle, fuzzy_match):
    '''Handle bundling for power and NC pins. Unbundle everything else.'''
    for unit in list(pin_data.values()):
        for side in list(unit.values()):
            for name, pins in list(side.items()):
                if len(pins) > 1:
                    for index, p in enumerate(pins):
                        if is_pwr(p, fuzzy_match) and bundle:
                            side[p.name + '_pwr'].append(p)
                        else:
                            side[p.name + '_' + str(index)].append(p)
                    del side[name]


def kipart(reader_type, part_data_file, parts_lib,
           allow_overwrite=False,
           sort_type='name',
           reverse=False,
           fuzzy_match=False,
           bundle=False,
           debug_level=0):
    '''Read part pin data from a CSV file and write or append it to a library file.'''

    part_reader_module = '{}_reader'.format(reader_type)
    importlib.import_module('kipart.' + part_reader_module)
    READER_MODULE = sys.modules['kipart.' + part_reader_module]
    part_reader = getattr(READER_MODULE, part_reader_module)

    # Get the part number and pin data from the CSV file.
    for part_num, part_ref_prefix, part_footprint, part_manf_num, pin_data in part_reader(part_data_file):

        # Handle retaining/overwriting parts that are already in the library.
        if parts_lib.get(part_num):
            if allow_overwrite:
                print('Overwriting part {}!'.format(part_num))
            else:
                print('Retaining previous definition of part {}.'.format(part_num))
                continue

        do_bundling(pin_data, bundle, fuzzy_match)

        # Draw the schematic symbol into the library.
        parts_lib[part_num] = \
                    draw_symbol(
                        part_num=part_num,
                        part_ref_prefix = part_ref_prefix,
                        part_footprint = part_footprint,
                        part_manf_num = part_manf_num,
                        pin_data=pin_data,
                        sort_type=sort_type,
                        reverse=reverse,
                        fuzzy_match=fuzzy_match)



# MIT license
# 
# Copyright (C) 2015 by XESS Corp.
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
import csv
import copy
import os.path
from collections import defaultdict
from .common import *
from .kipart import *


def lattice_reader(csv_file):
    '''Extract the pin data from a Lattice CSV file and return a dictionary of pin data.
    Both csv files available on Lattice website and csv files exported from Diamond are supported.
    ICE40 family is NOT supported, since they use a completely different format.'''

    # Create a dictionary that uses the package name as key. Each entry in this dictionary
    # uses the unit numbers as keys. Each entry in this dictionary contains another 
    # dictionary that uses the side of the symbol as a key. Each entry in that dictionary 
    # uses the pin names in that unit and on that side as keys. Each entry in that 
    # dictionary is a list of Pin objects with each Pin object having the same name
    # as the dictionary key. So the pins are separated into units at the top level, and then
    # the sides of the symbol, and then the pins with the same name that are on that side
    # of the unit.
    pin_data = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(list))))

    # Parse the first lines, until we find the column title row
    global_device_name=''
    global_package_name=''
    while True:
        pos=csv_file.tell()
        line = csv_file.readline()
        if line[0:8]=='#DEVICE=':
        global_device_name=line[9:].split(',')[0].strip().upper().replace('-', '_').replace(' ','_')
        if line[0:9]=='#PACKAGE=':
        global_package_name=line[10:].split(',')[0].strip().upper().replace('-', '_').replace(' ','_')
        test_field = line.split(',')[0].strip().upper()
        if test_field == 'INDEX' or test_field == 'PAD':
            break
    csv_file.seek(pos)

    # If no device name found in comments, get part number from file name
    if global_device_name:
        part_num = global_device_name
    else:
        part_num = os.path.splitext(os.path.basename(csv_file.name))[0]
    part_num = part_num.upper().split('PINOUT')[0].replace('-', '_').replace(' ','_')

    # Create a csv dict reader, from the column title row
    csv_reader = csv.DictReader(csv_file, skipinitialspace=True)

    # Some field name normalization : there is some variation in capitalization, 
    # 'Pin/Ball function' is sometimes 'Pin/Ball', and there may be references to comments after the field name (eg 'type(1)')
    csv_reader.fieldnames = [f.split('(')[0] for f in csv_reader.fieldnames]
    csv_reader.fieldnames = [f.strip().upper() for f in csv_reader.fieldnames]
    csv_reader.fieldnames = ['PIN/BALL FUNCTION' if f == 'PIN/BALL' else f for f in csv_reader.fieldnames]

    # If we have found a package name in comments, and a 'Pin Number' field is available,
    # we use this package name, and replace the 'Pin Number' name by its name
    # else we consider this is a multi-package csv file, and any column with an unknown field name is considered a package name
    # In the multi-package case, package names are normalized, and DQS variants are ignored
    if global_package_name and 'PIN NUMBER' in csv_reader.fieldnames:
        csv_reader.fieldnames = [global_package_name if f=='PIN NUMBER' else f for f in csv_reader.fieldnames]
        package = [global_package_name]
    else:
        known_fields = ['INDEX', 'TYPE', 'PAD', 'PIN/BALL FUNCTION', 'BANK', 'DUAL FUNCTION', 'DIFFERENTIAL', 'HIGH SPEED', 'DQS', 'I/O GROUPING', 'PIN NUMBER']
    csv_reader.fieldnames = [x if x in known_fields else x.upper().replace('-', '_').replace(' ','_') for x in csv_reader.fieldnames]
        package = [x for x in csv_reader.fieldnames if x and x not in known_fields and not x.endswith('_DQS')]

    if not package:
    print 'Warning : no package found, exiting'
        return

    #Process the pins line-by-line
    for index,row in enumerate(csv_reader):
        pin = copy.copy(DEFAULT_PIN)
        pin.index = index
        pin.name = row['PIN/BALL FUNCTION']
        if not row['BANK'] or row['BANK'] == '-' or row['BANK'] == ' ':
            pin.unit = 1
        else:
            pin.unit = int(row['BANK'])+2

        # The type of the pin isn't given in the text file, so we'll have to infer it
        # from the name of the pin. Pin names starting with the following prefixes 
        # are assigned the given pin type.
        DEFAULT_PIN_TYPE = 'io'  # Assign this pin type if name inference can't be made.
        PIN_TYPE_PREFIXES = [
            (r'VCC', 'power_in'),
            (r'GND', 'power_in'),
            (r'NC', 'no_connect'),
        ]
        for prefix, typ in PIN_TYPE_PREFIXES:
            if re.match(prefix, pin.name, re.IGNORECASE):
                pin.type = typ
                break
        else:
            pin.type = DEFAULT_PIN_TYPE

        # Same for pin side, in order to have VCC at the top and GND at the bottom
        PIN_SIDE_PREFIXES = [
            (r'VCC', 'top'),
            (r'GND', 'bottom'),
        ]
        for prefix, s in PIN_SIDE_PREFIXES:
            if re.match(prefix, pin.name, re.IGNORECASE):
                pin.side = s

        for p in package:
            if row[p] and row[p] != '-' and row[p] != ' ':
                pin.num = row[p]
                pin_data[p][pin.unit][pin.side][pin.name].append(copy.copy(pin))

        
    for p in package:
        yield part_num+'_'+p, 'U', '', part_num, pin_data[p]  # Return the dictionary of pins for the package p




# MIT license
# 
# Copyright (C) 2015 by XESS Corp.
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
import csv
import copy
import warnings
from collections import defaultdict
from .common import *
from .kipart import *

defaulted_names = set(list())

def xilinxultra_reader(csv_file):
    '''Extract the pin data from a Xilinx CSV file and return a dictionary of pin data.'''

    # Create a dictionary that uses the unit numbers as keys. Each entry in this dictionary
    # contains another dictionary that uses the side of the symbol as a key. Each entry in
    # that dictionary uses the pin names in that unit and on that side as keys. Each entry
    # in that dictionary is a list of Pin objects with each Pin object having the same name
    # as the dictionary key. So the pins are separated into units at the top level, and then
    # the sides of the symbol, and then the pins with the same name that are on that side
    # of the unit.
    pin_data = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))

    # Scan the initial portion of the file for the part number.
    part_num = None
    try:
        while True:
            line = csv_file.readline()
            if re.match('^,*$', line):
                # Stop searching for part number as soon as a blank line is seen.
                break
            elif line.startswith('"#') or line.startswith('#'):
                # Look for the part number within a comment.
                device = re.search(r'#\s+Device\s*:\s*(\w+)', line)
                if device:
                    part_num = device.group(1)
            else:
                # Look for the part number on a line of the file.
                _, part_num, date, time, _ = re.split('\s+', line)
    except Exception:
        return  # No part was found.

    if part_num is None:
        return  # No part number was found, so abort.

    # Create a reader object for the rows of the CSV file and read it row-by-row.
    csv_reader = csv.DictReader(csv_file, skipinitialspace=True)
    for index, row in enumerate(csv_reader):
        # A blank line signals the end of the pin data.
        try:
            if row['Pin'] == '':
                break
        except KeyError:
            # Abort if a TXT file is being processed instead of a CSV file.
            return

        # Get the pin attributes from the cells of the row of data.
        pin = copy.copy(DEFAULT_PIN)
        pin.index = index
        pin.name = fix_pin_data(row['Pin Name'], part_num)
        pin.num = fix_pin_data(row['Pin'], part_num)
        pin.unit = fix_pin_data(row['Bank'], part_num)

        # The type of the pin isn't given in the CSV file, so we'll have to infer it
        # from the name of the pin. Pin names starting with the following prefixes 
        # are assigned the given pin type.
        DEFAULT_PIN_TYPE = 'input'  # Assign this pin type if name inference can't be made.
        PIN_TYPE_PREFIXES = [
            (r'CCLK', 'bidirectional'),
            (r'CFGBVS_', 'input'),
            (r'DONE', 'bidirectional'),
            (r'D0[0-3]_', 'bidirectional'),
            (r'DXP', 'passive'),
            (r'DXN', 'passive'),
            (r'GNDADC', 'input'),
            (r'GND', 'power_in'),
            (r'RSVDGND', 'input'),
            (r'PUDC_B', 'input'),
            (r'INIT_B', 'bidirectional'),
            (r'IO_', 'bidirectional'),
            (r'M0[_]?', 'input'),
            (r'M1[_]?', 'input'),
            (r'M2[_]?', 'input'),
            (r'MGTAVCC[_]?', 'power_in'),
            (r'MGTAVTTRCAL_', 'input'),
            (r'MGTAVTT[_]?', 'input'),
            (r'MGTHRX[NP][0-9]+_', 'input'),
            (r'MGTHTX[NP][0-9]+_', 'output'),
            (r'MGTREFCLK[0-9][NP]+_', 'input'),
            (r'MGTRREF_', 'input'),
            (r'MGTVCCAUX[_]?', 'power_in'),
            (r'MGTYRX[NP][0-9]+_', 'input'),
            (r'MGTYTX[NP][0-9]+_', 'output'),
            (r'NC', 'no_connect'),
            (r'POR_OVERRIDE', 'input'),
            (r'PUDC_B_[0-9]+', 'input'),
            (r'PROGRAM_B_[0-9]+', 'input'),
            (r'RDWR_FCS_B_[0-9]+', 'bidirectional'),
            (r'TCK_[0-9]+', 'input'),
            (r'TDI_[0-9]+', 'input'),
            (r'TDO_[0-9]+', 'output'),
            (r'TMS_[0-9]+', 'input'),
            (r'VBATT', 'power_in'),
            (r'VCCADC?', 'power_in'),
            (r'VCCAUX[_]?', 'power_in'),
            (r'VCCBRAM', 'power_in'),
            (r'VCCINT', 'power_in'),
            (r'VCCO_', 'power_in'),
            (r'VN', 'input'),
            (r'VP', 'input'),
            (r'VREF[PN]', 'input'),
            (r'VREF_', 'input'),

            (r'PS_MIO[0-9]+', 'bidirectional'),
            (r'PS_DDR_DQ[0-9]+', 'bidirectional'),
            (r'PS_DDR_DQS_[PN][0-9]+', 'bidirectional'),
            (r'PS_DDR_ALERT_N', 'input'),
            (r'PS_DDR_ACT_N', 'output'),
            (r'PS_DDR_A[0-9]+', 'output'),
            (r'PS_DDR_BA[0-9]+', 'output'),
            (r'PS_DDR_BG[0-9]+', 'output'),
            (r'PS_DDR_CK_N[0-9]+', 'output'),
            (r'PS_DDR_CK[0-9]+', 'output'),
            (r'PS_DDR_CKE[0-9]+', 'output'),
            (r'PS_DDR_CS_N[0-9]+', 'output'),
            (r'PS_DDR_DM[0-9]+', 'output'),
            (r'PS_DDR_ODT[0-9]+', 'output'),
            (r'PS_DDR_PARITY[0-9]*', 'output'),
            (r'PS_DDR_RAM_RST_N[0-9]*', 'output'),
            (r'PS_DDR_ZQ[0-9]*', 'bidirectional'),
            (r'VCC_PS', 'power_in'),
            (r'PS_DONE', 'output'),
            (r'PS_ERROR_OUT', 'output'),
            (r'PS_ERROR_STATUS', 'output'),
            (r'PS_MODE[0-9]+', 'input'),
            (r'PS_PADI', 'input'),
            (r'PS_PADO', 'output'),
            (r'PS_POR_B', 'input'),
            (r'PS_PROG_B', 'input'),
            (r'PS_INIT_B', 'output'),
            (r'PS_DONE', 'output'),
            (r'PS_REF_CLK', 'input'),
            (r'PS_SRST_B', 'input'),
            (r'PS_MGTRRX[NP][0-9]+_', 'input'),
            (r'PS_MGTRTX[NP][0-9]+_', 'output'),
            (r'PS_MGTREFCLK[0-9]+[NP]_', 'input'),
            (r'PS_MGTRAVCC', 'power_in'),
            (r'PS_MGTRAVTT', 'power_in'),
            (r'PS_MGTRREF', 'input'),
            (r'PS_JTAG_TCK', 'input'),
            (r'PS_JTAG_TDI', 'input'),
            (r'PS_JTAG_TDO', 'output'),
            (r'PS_JTAG_TMS', 'input'),
        ]
        for prefix, typ in PIN_TYPE_PREFIXES:
            if re.match(prefix, pin.name, re.IGNORECASE):
                pin.type = typ
                break
        else:
            if pin.name not in defaulted_names:
                warnings.warn('No match for {} on {}, assigning as {}'.format(
                    pin.name, part_num[:4], DEFAULT_PIN_TYPE))
            pin.type = DEFAULT_PIN_TYPE
            defaulted_names.add(pin.name)
        pin.type = fix_pin_data(pin.type, part_num)

        # Add the pin from this row of the CVS file to the pin dictionary.
        # Place all the like-named pins into a list under their common name.
        # We'll unbundle them later, if necessary.
        pin_data[pin.unit][pin.side][pin.name].append(pin)

    yield part_num, 'U', '', part_num, pin_data  # Return the dictionary of pins extracted from the CVS file.



# MIT license
#
# Copyright (C) 2016 Hasan Yavuz Ozderya
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import csv
from collections import defaultdict
import re
from operator import itemgetter
import os
import copy
from collections import defaultdict
from .common import *
from .kipart import *

# Pin type mappings of STM32Cube output to kipart accepted values.
type_mappings = {
    "Power" : "power_in",
    "Input" : "input",
    "Output" : "output",
    "I/O" : "inout",
    "Reset" : "input",
    "Boot" : "input"
}

def parse_csv_file(csv_file):
    """Parses the CSV file and returns a list of pins in the form of (number, 'name', 'type')"""

    pins = []
    reader = csv.reader(csv_file, delimiter=',', quotechar='"')
    next(reader, None) # skip header
    for row in reader:
        number, name, ptype, signal, label = row

        if label:
            name += '/' + label
        elif signal:
            name += '/' + signal

        name = name.replace(' ', '_')
        pin_type = type_mappings.get(ptype, "inout")

        pins.append((number, name, pin_type))

    return pins

def parse_portpin(name):
    """Finds the port name and number of a pin in a string. If found
    returns a tuple in the form of ('port_name', port_number).
    Otherwise returns `None`.
    """
    m = re.search('P([A-Z])(\d+)', name)
    if m:
        port_name, port_number = m.groups()
        return (port_name, int(port_number))

def group_pins(pins):
    """Groups pins together per their port name and functions. Returns a
    dictionary of {'port': [pin]}."""
    ports = defaultdict(list)

    power_names = ['VDD', 'VSS', 'VCAP', 'VBAT', 'VREF']
    config_names = ['RCC_OSC', 'NRST', 'PDR', 'SWCLK', 'SWDIO', 'BOOT']

    for pin in pins:
        number, name, ptype = pin
        if any(pn in name for pn in power_names):
            ports['power'].append(pin)

        elif any(pn in name for pn in config_names):
            ports['config'].append(pin)

        else:
            m = parse_portpin(name)
            if m:
                port_name, port_number = m
                ports[port_name].append(pin)
            else:
                ports['other'].append(pin)

    # sort pins
    for port in ports:
        # config and power gates are sorted according to their function name
        if port in ['config', 'power']:
            ports[port] = sorted(ports[port], key=itemgetter(1))
        # IO ports are sorted according to port number
        else:
            ports[port] = sorted(ports[port], key=lambda p: parse_portpin(p[1])[1])

    return ports

def stm32cube_reader(csv_file):
    """Reader for STM32CubeMx pin list output.

    STM32CubeMx is a tool for creating firmware projects for STM32
    MCUs. It also includes a pin layout designer which can export the
    list of pins in the form of a CSV file. This will read the csv
    file and return a dictionary of pin data.

    An example output of the STM32CubeMx tool can be seen below:

    "Position","Name","Type","Signal","Label"
    "1","VBAT","Power","",""
    "2","PC13-ANTI_TAMP","I/O","",""
    "3","PC14-OSC32_IN","I/O","RCC_OSC32_IN",""
    "4","PC15-OSC32_OUT","I/O","RCC_OSC32_OUT",""
    ...

    Pin names for the symbols will be constructed as "Name/Signal". If
    user defined label is specified it will be used instead of
    "Signal" column.

    All IO pins will be grouped to units per their ports.
    Configuration related pins such as boot, clock etc will be grouped
    as a separate unit. Power pins will be a separate unit as well.
    """
    pins = parse_csv_file(csv_file)

    ports = group_pins(pins)

    # create pin data
    pin_data = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))

    index = 0

    for port_name in ports:
        for p in ports[port_name]:
            # Get the pin attributes from the cells of the row of data.
            pin = copy.copy(DEFAULT_PIN) # Start off with default values for the pin.
            pin.index = index = index + 1
            pin.num = p[0]
            pin.name = p[1]
            pin.type = p[2]
            pin.unit = port_name

            pin_data[pin.unit][pin.side][pin.name].append(pin)

    # use file name as the part name
    part_name = os.path.splitext(os.path.split(csv_file.name)[1])[0]

    # what should be the part_num?
    yield part_name, 'U', '', '', pin_data



# MIT license
# 
# Copyright (C) 2016 by XESS Corporation.
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.



"""
Parsers for netlist files of various formats (only KiCad, at present).
"""


from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import open
from future import standard_library
standard_library.install_aliases()

from random import randint, choice
import string
import re
import sys

THIS_MODULE = locals()

def gen_random_part_csv():
    def random_name(len=20):
        chars = string.printable[:94]
        chars = re.sub('"','',chars)
        chars = re.sub("'",'',chars)
        chars = re.sub(',','',chars)
        chars = re.sub('\|','',chars)
        return ''.join(choice(chars) for _ in range(randint(1,len)))

    def zero_pad_nums(s):
        # Pad all numbers in the string with leading 0's.
        # Thus, 'A10' and 'A2' will become 'A00010' and 'A00002' and A2 will
        # appear before A10 in a list.
        try:
            return re.sub(r'\d+', lambda mtch: '0' * (8 - len(mtch.group(0))) + mtch.group(0), s)
        except TypeError:
            return s # The input is probably not a string, so just return it unchanged.

    def pin_key(unit_pinnum_tuple):
        return zero_pad_nums(unit_pinnum_tuple[0]) + zero_pad_nums(unit_pinnum_tuple[1])

    part_name = random_name()
    ref_prefix = random_name(2)
    part_csv = '{part_name},{ref_prefix},,,,,'.format(**locals()) 
    part_csv = '\n'.join([part_csv, 'Pin,Name,Type,Side,Unit,Style,Hidden'])

    num_units = randint(1,10)
    pins = {}
    for unit in range(1, num_units+1):
        unit_name = str(unit)
        for _ in range(randint(num_units,256)//num_units):
            pin_num = (choice(string.ascii_uppercase+'           ') + str(randint(1,256))).strip()
            pin_name = random_name()
            pin_side = choice(['left','right','top','bottom'])
            pin_type = choice(['in','out','bidir','tri','passive','pwr','pwr_out','open_collector','open_emitter','unspecified','NC'])
            pin_style = choice(['','inv','clk', 'inv_clk', 'input_low', 'clk_low', 'output_low', 'falling_clk', 'non_logic'])
            pin_hidden = choice(['Y',''])
            pin_csv = '{pin_num},{pin_name},{pin_type},{pin_side},{unit_name},{pin_style},{pin_hidden}'.format(**locals())
            pins[(unit_name,pin_num)] = pin_csv
    for p in sorted(pins.keys(),key=pin_key):
        part_csv = '\n'.join([part_csv,pins[p]])

    part_csv = '\n'.join([part_csv,',,,,,,'])
    return part_csv

def gen_random_lib_csv(num_lines=1000):
    lib_csv = gen_random_part_csv()
    while lib_csv.count('\n') < num_lines:
        lib_csv = '\n'.join([lib_csv, gen_random_part_csv()])
    lib_csv += '\n'
    return lib_csv

if __name__ == '__main__':
    print(gen_random_lib_csv())



# MIT license
#
# Copyright (C) 2015 by XESS Corp.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function
from builtins import object
import re
import difflib

COLUMN_NAMES = {
    'pin': 'num',
    'num': 'num',
    'name': 'name',
    'type': 'type',
    'style': 'style',
    'side': 'side',
    'unit': 'unit',
    'bank': 'unit',
    'hidden': 'hidden'
}


# This is just a vanilla object class for device pins.
# We'll add attributes to it as needed.
class Pin(object):
    pass


DEFAULT_PIN = Pin()
DEFAULT_PIN.num = None
DEFAULT_PIN.name = ''
DEFAULT_PIN.type = 'io'
DEFAULT_PIN.style = 'line'
DEFAULT_PIN.unit = 1
DEFAULT_PIN.side = 'left'
DEFAULT_PIN.hidden = 'no'


def num_row_elements(row):
    '''Get number of elements in CSV row.'''
    try:
        rowset = set(row)
        rowset.discard('')
        return len(rowset)
    except TypeError:
        return 0


def get_nonblank_row(csv_reader):
    '''Return the first non-blank row encountered from the current point in a CSV file.'''
    for row in csv_reader:
        if num_row_elements(row) > 0:
            return row
    return []


def get_part_info(csv_reader):
    '''Get the part number, ref prefix, footprint, MPN from a row of the CSV file.'''

    # Read the first, nonblank row and pad it with None's to make sure it's long enough.
    part_num, part_ref_prefix, part_footprint, part_manf_num = list(get_nonblank_row(csv_reader) + [None]*4)[:4]

    # Put in the default part reference identifier if it isn't present.
    if part_ref_prefix in (None, '', ' '):
        part_ref_prefix = 'U'

    # Check to see if the row with the part identifier is missing.
    if part_num and part_num.lower() in list(COLUMN_NAMES.keys()):
        issue('Row with part number is missing in CSV file.', 'error')

    return part_num, part_ref_prefix, part_footprint, part_manf_num


def find_closest_match(name, name_dict, fuzzy_match, threshold=0.0):
    '''Approximate matching subroutine'''
    # Scrub non-alphanumerics from name and lowercase it.
    scrubber = re.compile('[\W.]+')
    name = scrubber.sub('', name).lower()

    # Return regular dictionary lookup if fuzzy matching is not enabled.
    if fuzzy_match == False:
        return name_dict[name]

    # Find the closest fuzzy match to the given name in the scrubbed list.
    # Set the matching threshold to 0 so it always gives some result.
    match = difflib.get_close_matches(name, list(name_dict.keys()), 1, threshold)[0]

    return name_dict[match]


def clean_headers(headers):
    '''Return a list of the closest valid column headers for the headers found in the file.'''
    return [find_closest_match(h, COLUMN_NAMES, True) for h in headers]


def issue(msg, level='warning'):
    if level=='warning':
        print('Warning: {}'.format(msg))
    elif level == 'error':
        print('ERROR: {}'.format(msg))
        raise Exception('Unrecoverable error')
    else:
        print(msg)


def fix_pin_data(pin_data, part_num):
    '''Fix common errors in pin data.'''
    fixed_pin_data = pin_data.strip()  # Remove leading/trailing spaces.
    if re.search('\s', fixed_pin_data) is not None:
        fixed_pin_data = re.sub('\s', '_', fixed_pin_data)
        issue("Replaced whitespace with '_' in pin '{pin_data}' of part {part_num}.".format(**locals()))
    return fixed_pin_data




# MIT license
#
# Copyright (C) 2015 by XESS Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function
from __future__ import absolute_import
import argparse as ap
import os
import sys
import io
import zipfile
from collections import OrderedDict
from .__init__ import __version__
from .kipart import *
from .common import DEFAULT_PIN, issue

def main():
    parser = ap.ArgumentParser(
        description=
        'Generate single & multi-unit schematic symbols for KiCad from a CSV file.')

    parser.add_argument(
        '-v', '--version',
        action='version',
        version='KiPart ' + __version__)
    parser.add_argument(
        'input_files',
        nargs='+',
        type=str,
        metavar='file1.[csv|zip] file2.[csv|zip] ...',
        help='Files for parts in CSV format or as CSV files in .zip archives.')
    parser.add_argument(
        '-r', '--reader',
        nargs='?',
        type=str.lower,
        choices=['generic', 'xilinxultra', 'xilinx7', 'xilinx6s', 'xilinx6v', 'psoc5lp', 'stm32cube', 'lattice'],
        default='generic',
        help='Name of function for reading the CSV files.')
    parser.add_argument(
        '-s', '--sort',
        nargs='?',
        type=str.lower,
        choices=['row', 'num', 'name'],
        default='row',
        help=
        'Sort the part pins by their entry order in the CSV file, their pin number, or their pin name.')
    parser.add_argument(
        '--reverse',
        action = 'store_true',
        help='Sort pins in reverse order.'
    )
    parser.add_argument(
        '--side',
        nargs='?',
        type=str.lower,
        choices=['left', 'right', 'top', 'bottom'],
        default='left',
        help='Which side to place the pins by default.'
    )
    parser.add_argument(
        '-o', '--output',
        nargs='?',
        type=str,
        metavar='file.lib',
        help='Generated KiCad symbol library for parts.')
    parser.add_argument(
        '-f', '--fuzzy_match',
        action='store_true',
        help=
        'Use approximate string matching when looking-up the pin type, style and orientation.')
    parser.add_argument(
        '-b', '--bundle',
        action='store_true',
        help=
        'Bundle multiple, identically-named power, ground and no-connect pins each into a single schematic pin.')
    parser.add_argument(
        '-a', '--append', '--add',
        action='store_true',
        help='Add parts to an existing part library. Overwrite existing parts only if used in conjunction with -w.')
    parser.add_argument(
        '-w', '--overwrite',
        action='store_true',
        help='Allow overwriting of an existing part library.')
    parser.add_argument(
        '-d', '--debug',
        nargs='?',
        type=int,
        default=0,
        metavar='LEVEL',
        help='Print debugging info. (Larger LEVEL means more info.)')

    args = parser.parse_args()

    if args.output == None:
        issue('No output file specified. Using kipart.lib as the default.')
        args.output = 'kipart.lib'
    else:
        args.output = os.path.splitext(args.output)[0] + '.lib'

    import re
    parts_lib = OrderedDict()
    if os.path.isfile(args.output):
        if not args.overwrite and not args.append:
            print('Output file {} already exists! Use the --overwrite option to replace it or the --append option to append to it.'.format(
                args.output))
            sys.exit(1)
        if args.append:
            with open(args.output, 'r') as lib:
                part_def = ''
                for line in lib:
                    start = re.match('DEF (?P<part_name>\S+)', line)
                    end = re.match('ENDDEF$', line)
                    if start:
                        part_def = line
                        part_name = start.group('part_name')
                    elif end:
                        part_def += line
                        parts_lib[part_name] = part_def
                    else:
                        part_def += line

    def call_kipart(part_data_file):
        '''Helper routine for calling kipart.'''
        return kipart(reader_type=args.reader,
                   part_data_file=part_data_file,
                   parts_lib=parts_lib,
                   allow_overwrite=args.overwrite,
                   sort_type=args.sort,
                   reverse=args.reverse,
                   fuzzy_match=args.fuzzy_match,
                   bundle=args.bundle,
                   debug_level=args.debug)

    DEFAULT_PIN.side = args.side

    for input_file in args.input_files:
        file_ext = os.path.splitext(input_file)[-1]
        if file_ext == '.zip':
            zip_file = zipfile.ZipFile(input_file, 'r')
            zipped_files = zip_file.infolist()
            for zipped_file in zipped_files:
                if os.path.splitext(zipped_file.filename)[-1] in ['.csv', '.txt']:
                    with zip_file.open(zipped_file, 'r') as part_data_file:
                        part_data_file = io.TextIOWrapper(part_data_file)
                        call_kipart(part_data_file)
        elif file_ext in ['.csv', '.txt']:
            with open(input_file, 'r') as part_data_file:
                call_kipart(part_data_file)
        else:
            continue

    LIB_HEADER = 'EESchema-LIBRARY Version 2.3\n'
    with open(args.output, 'w') as lib_file:
        lib_file.write(LIB_HEADER)
        for part_num, part_defn in parts_lib.items():
            lib_file.write(part_defn)

# main entrypoint.
if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import setuptools
import kipart

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read().replace('.. :changelog:', '')

requirements = ['affine >= 1.2.0', 'future >= 0.15.0', 'pyparsing']

test_requirements = [  # TODO: put package test requirements here
]

setup(name='kipart',
      version=kipart.__version__,
      description="Part creator for KiCad.",
      long_description=readme + '\n\n' + history,
      author=kipart.__author__,
      author_email=kipart.__email__,
      url='https://github.com/xesscorp/kipart',
      packages=setuptools.find_packages(),
      entry_points={'console_scripts': 
            ['kipart = kipart.__main__:main', 
             'kilib2csv = kipart.kilib2csv:main']},
      package_dir={'kipart': 'kipart'},
      include_package_data=True,
      package_data={'kipart': ['*.gif', '*.png']},
      scripts=[],
      install_requires=requirements,
      license="MIT",
      zip_safe=False,
      keywords='kipart',
      classifiers=['Development Status :: 2 - Pre-Alpha',
                   'Intended Audience :: Developers',
                   'License :: OSI Approved :: MIT License',
                   'Natural Language :: English',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3.4', ],
      test_suite='tests',
      tests_require=test_requirements)



# MIT license
#
# Copyright (C) 2016 by XESS Corporation.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


"""
Parsers for LiCad schematic symbol library files.
"""



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import open
from future import standard_library
standard_library.install_aliases()

import argparse as ap
import os
import sys
import re
from pyparsing import *

from .py_2_3 import *

from .__init__ import __version__

THIS_MODULE = locals()

def _parse_lib_kicad(text):
    """
    Return a pyparsing object storing the contents of a KiCad symbol library.
    """

    # Basic parser elements.
    terminator_char = '|'
    terminator = Literal(terminator_char + '\n').suppress()
    fnum = Group(Word(nums) + Optional(Literal('.') + Optional(Word(nums))))
    string = Word(printables, excludeChars=terminator_char)
    qstring = dblQuotedString() ^ sglQuotedString()
    qstring.addParseAction(removeQuotes)
    anystring = qstring ^ string

    #------------------ Schematic symbol parser.--------------------

    # Field section parser. (Fields are ignored.)
    field_start = CaselessLiteral('F') + Word(nums)('field_num')
    field = field_start + restOfLine
    fields = ZeroOrMore(field).suppress()  # Just ignore all fields.

    # Part aliases. (Aliases are ignored.)
    aliases = ZeroOrMore(CaselessKeyword('ALIAS') + restOfLine()).suppress()

    # Footprint section. (Footprints are just skipped over and ignored.)
    foot_start = CaselessKeyword('$FPLIST') + terminator
    foot_end = CaselessKeyword('$ENDFPLIST') + terminator
    footprints = Optional(foot_start + SkipTo(foot_end) + foot_end).suppress()

    # Draw section parser.
    draw_start = (CaselessKeyword('DRAW') + restOfLine).suppress()
    draw_end = (CaselessKeyword('ENDDRAW') + terminator).suppress()
    draw_pin = Group(Word('Xx', exact=1).suppress() + anystring('name') + anystring('num') + \
                (anystring * 3).suppress() + anystring('orientation') + \
                (anystring * 2).suppress() + anystring('unit') + anystring.suppress() + \
                anystring('type') + Optional(anystring)('style') + terminator)
    draw_other = (Word('AaCcPpSsTt', exact=1) + White() + restOfLine).suppress()
    draw_element = draw_pin ^ draw_other
    draw = Group(draw_start + ZeroOrMore(draw_element) + draw_end)('pins')

    # Complete schematic symbol definition parser.
    def_start = CaselessKeyword('DEF').suppress() + anystring('name') + anystring('ref_id') + restOfLine.suppress()
    def_end = (CaselessKeyword('ENDDEF') + terminator).suppress()
    defn = Group(def_start + (fields & aliases & footprints & draw) + def_end)

    # Parser for set of schematic symbol definitions.
    defs = ZeroOrMore(defn)

    # Parser for entire library.
    version = CaselessKeyword('version').suppress() + fnum
    start = CaselessKeyword('EESchema-LIBRARY').suppress() + \
            SkipTo(version) + version('version') + restOfLine.suppress()
    end_of_file = Optional(White()) + stringEnd
    lib = start + defs('parts') + end_of_file

    #---------------------- End of parser -------------------------

    # Remove all comments from the text to be parsed but leave the lines blank.
    # (Don't delete the lines or else it becomes hard to find the line in the file
    # that made the parser fail.)
    text = re.sub('(^|\n)#.*?(?=\n)', '\n', text)

    # Terminate all lines with the terminator string.
    # (This makes it easier to handle each line without accidentally getting
    # stuff from the next line.)
    text = re.sub('\n', ' '+terminator_char + '\n', text)

    # Remove the terminator from all lines that just have the terminator character on them.
    # (Don't delete the lines or else it becomes hard to find the line in the file
    # that made the parser fail.)
    text = re.sub('\n *\\' + terminator_char + ' *(?=\n)', '\n', text)

    # Return the parsed text.
    return lib.parseString(text)


def _parse_lib(src, tool='kicad'):
    """
    Return a pyparsing object storing the contents of a schematic symbol library.

    Args:
        src: Either a text string, or a filename, or a file object that stores
            the netlist.

    Returns:
        A pyparsing object that stores the library contents.

    Exception:
        PyparsingException.
    """

    try:
        text = src.read()
    except Exception:
        try:
            text = open(src,'r').read()
        except Exception:
            text = src

    if not isinstance(text, basestring):
        raise Exception("What is this shit you're handing me? [{}]\n".format(src))

    try:
        # Use the tool name to find the function for loading the library.
        func_name = '_parse_lib_{}'.format(tool)
        parse_func = THIS_MODULE[func_name]
        return parse_func(text)
    except KeyError:
        # OK, that didn't work so well...
        logger.error('Unsupported ECAD tool library: {}'.format(tool))
        raise Exception

def _gen_csv(parsed_lib):
    '''Return multi-line CSV string for the parts in a parsed schematic library.'''

    type_tbl = {'I':'in', 'O':'out', 'B':'bidir', 'T':'tri', 'P':'passive', 'U':'unspecified', 'W':'pwr', 'w':'pwr_out', 'C':'open_collector', 'E':'open_emitter', 'N':'NC'}
    orientation_tbl = {'R':'left', 'L':'right', 'U':'bottom', 'D':'top'}
    style_tbl = {'':'', 'I':'inv', 'C':'clk', 'IC':'inv_clk', 'L':'input_low', 'CL':'clk_low', 'V':'output_low', 'F':'falling_clk', 'X':'non_logic'}

    csv = ''
    for part in parsed_lib.parts:
        csv += '{part.name},{part.ref_id},,,,,\n'.format(**locals())
        csv += 'Pin,Name,Type,Side,Unit,Style,Hidden\n'

        def zero_pad_nums(s):
            # Pad all numbers in the string with leading 0's.
            # Thus, 'A10' and 'A2' will become 'A00010' and 'A00002' and A2 will
            # appear before A10 in a list.
            try:
                return re.sub(r'\d+', lambda mtch: '0' * (8 - len(mtch.group(0))) + mtch.group(0), s)
            except TypeError:
                return s # The input is probably not a string, so just return it unchanged.

        def num_key(pin):
            '''Generate a key from a pin's number so they are sorted by position on the package.'''

            # Pad all numeric strings in the pin name with leading 0's.
            # Thus, 'A10' and 'A2' will become 'A00010' and 'A00002' and A2 will
            # appear before A10 in a list.
            return zero_pad_nums(pin.unit) + zero_pad_nums(pin.num)

        for p in sorted(part.pins, key=num_key):
            # Replace commas in pin numbers, names and units so it doesn't screw-up the CSV file.
            p.num = re.sub(',', ';', p.num)
            p.name = re.sub(',', ';', p.name)
            p.unit = re.sub(',', ';', p.unit)

            is_hidden = ''

            if p.style.find('N') != -1:
                p.style = p.style.replace('N', '')
                is_hidden = 'Y'

            csv += ','.join([p.num, p.name, type_tbl[p.type], orientation_tbl[p.orientation], p.unit, style_tbl[p.style], is_hidden]) + '\n'
        csv += ',,,,,,\n'
    csv += '\n'
    return csv

def main():
    parser = ap.ArgumentParser(
        description=
        'Convert a KiCad schematic symbol library file into a CSV file for KiPart.')

    parser.add_argument('-v', '--version',
        action='version',
        version='kilib2csv ' + __version__)
    parser.add_argument(
        'input_files',
        nargs='+',
        type=str,
        metavar='file.lib',
        help='KiCad schematic symbol library.')
    parser.add_argument('-o', '--output',
        nargs='?',
        type=str,
        metavar='file.csv',
        help='CSV file created from schematic library file.')
    parser.add_argument('-a', '--append',
        action='store_true',
        help='Append to an existing CSV file.')
    parser.add_argument('-w', '--overwrite',
        action='store_true',
        help='Allow overwriting of an existing CSV file.')

    args = parser.parse_args()

    if args.output == None:
        issue('No output file specified. Using kipart.csv as the default.')
        args.output = 'kipart.csv'
    else:
        args.output = os.path.splitext(args.output)[0] + '.csv'

    if os.path.isfile(args.output):
        if not (args.overwrite or args.append):
            print('Output file {} already exists! Use the --overwrite option to replace it or the --append option to append to it.'.format(
                args.output))
            sys.exit(1)

    file_mode = 'a' if args.append else 'w'
    with open(args.output, file_mode) as out_file:
        for lib in args.input_files:
            parsed_lib = _parse_lib(lib)
            out_file.write(_gen_csv(parsed_lib))

# main entrypoint.
if __name__ == '__main__':
    main()

